# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Provide a component to handle the summary view for system and actor node types."""

from dash import dash_table
import dash_bootstrap_components as dbc
from dash_extensions.enrich import dcc, html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.actor_information_table import ActorInformationTable
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.actor_logs_table import ActorLogsTable
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.actor_statistics_table import ActorStatisticsTable
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.system_files import SystemFiles


class SummaryView:
    """Summary view component."""

    def __init__(self) -> None:
        """Constructor."""

        self.actor_info: dict = None
        self.actor_status_info: dict = None
        self.result_files: dict = None
        self.font_size: str = "15px"

    def _get_actor_information_table(self) -> dash_table.DataTable:

        table = ActorInformationTable()

        if self.actor_info:
            table.actor_info = self.actor_info
        if self.actor_status_info:
            table.actor_status_info = self.actor_status_info

        return table.render()

    def _get_system_files(self) -> html.Div:

        system_files = SystemFiles()

        if self.actor_info:
            system_files.node_name = self.actor_info["name"]
        if self.result_files:
            system_files.omdb_file = self.result_files["omdb"]

        return system_files.render()

    def _get_actor_logs_table(self) -> dash_table.DataTable:

        table = ActorLogsTable()

        if self.actor_info:
            table.actor_info = self.actor_info

        return table.render()

    def _get_actor_statistics_table(self) -> dash_table.DataTable:

        table = ActorStatisticsTable()

        if self.actor_info:
            table.actor_info = self.actor_info

        return table.render()

    def render(self):
        """Generate layout."""

        # Actor information ---------------------------------------------------

        actor_information_card = dbc.Card(
            [
                dbc.CardBody(
                    [
                        html.H4("Actor information", className="card-title"),
                        self._get_actor_information_table(),
                    ]
                ),
            ],
            color="warning",
            outline=True,
        )

        # Actor commands ------------------------------------------------------

        actor_commands_card = dbc.Card(
            [
                dbc.CardBody(
                    [
                        html.H4("Actor commands", className="card-title"),
                        dcc.Markdown("🚧 Under construction!", style={"font-size": "20px"}),
                    ]
                ),
            ],
            color="warning",
            outline=True,
        )

        # System Files --------------------------------------------------------

        system_files_card = dbc.Card(
            [
                dbc.CardBody(
                    [
                        html.H4("System Files", className="card-title"),
                        self._get_system_files(),
                    ]
                ),
            ],
            color="warning",
            outline=True,
        )

        # Actor log -----------------------------------------------------------

        actor_log_card = dbc.Card(
            [
                dbc.CardBody(
                    [
                        html.H4("Actor Log", className="card-title"),
                        self._get_actor_logs_table(),
                    ]
                ),
            ],
            color="warning",
            outline=True,
        )

        # Actor statistics -----------------------------------------------------------

        actor_statistics_card = dbc.Card(
            [
                dbc.CardBody(
                    [
                        html.H4("Actor Statistics", className="card-title"),
                        self._get_actor_statistics_table(),
                    ]
                ),
            ],
            color="warning",
            outline=True,
        )

        # Assembly -------------------------------------------------------------------

        content = [
            html.Br(),
            dbc.Row(
                [
                    dbc.Col(actor_information_card, width=12),
                ]
            ),
            html.Br(),
            dbc.Row(
                [
                    dbc.Col(actor_commands_card, width=12),
                ]
            ),
        ]

        if self.actor_info:
            if self.actor_info["kind"] == "system":
                content.extend(
                    [
                        html.Br(),
                        dbc.Row(
                            [
                                dbc.Col(system_files_card, width=12),
                            ]
                        ),
                    ]
                )

        content.extend(
            [
                html.Br(),
                dbc.Row(
                    [
                        dbc.Col(actor_log_card, width=12),
                    ]
                ),
                html.Br(),
                dbc.Row(
                    [
                        dbc.Col(actor_statistics_card, width=12),
                    ]
                ),
            ]
        )

        return html.Div(content)



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Provide a component to handle the design table for root and system node types."""

from dash import dash_table
import pandas as pd

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.utils.common_functions import sorted_nicely


class DesignTable:
    """Design table component."""

    def __init__(self) -> None:
        """Constructor."""

        self.actor_info: dict = None
        self.actor_status_info: dict = None
        self._field_types = ["parameter", "constraint", "objective", "response"]
        self.font_size: str = "15px"

    def _sort_designs(self, designs: list) -> list:
        """"""

        sorted_designs = sorted_nicely(designs)
        return [designs.index(design) for design in sorted_designs]

    def _get_data(self) -> pd.DataFrame:

        design_table_data = {
            "Design": [],
            "Feasible": [],
            "Status": [],
            "Pareto": [],
        }

        if self.actor_info:
            if self.actor_info["kind"] == "system":
                for field_type in self._field_types:
                    field_name_index = field_type + "_names"
                    for field_name in self.actor_status_info["designs"][field_name_index]:
                        design_table_data[field_name] = []
                design_numbers = []
                for design_values in reversed(self.actor_status_info["designs"]["values"]):
                    design_numbers.append(design_values["hid"].split(".")[-1])
                    design_table_data["Design"].append(design_values["hid"])
                    for field_type in self._field_types:
                        field_name_index = field_type + "_names"
                        field_value_index = field_type + "_values"
                        field_names = self.actor_status_info["designs"][field_name_index]
                        field_values = design_values[field_value_index]
                        if len(field_names) != len(field_values):
                            field_values = [None for i in range(len(field_names))]
                        for field_name, field_value in zip(field_names, field_values):
                            if isinstance(field_value, dict):
                                if field_value["type"] == "xy_data":
                                    field_value = f"[1:%s]" % (field_value["num_entries"])
                            design_table_data[field_name].append(field_value)
                for design_status in reversed(self.actor_status_info["design_status"]):
                    design_table_data["Feasible"].append(design_status["feasible"])
                    design_table_data["Status"].append(design_status["status"])
                    design_table_data["Pareto"].append(design_status["pareto_design"])

                design_table_data = pd.DataFrame(design_table_data)

                # OptiSLang returns the designs in an unsorted way.
                # The following method organizes the rows in ascending order.
                design_table_data = design_table_data.loc[self._sort_designs(design_table_data["Design"].to_list())]

                for col, dtype in design_table_data.dtypes.items():
                    if dtype == "bool":
                        design_table_data[col] = design_table_data[col].astype("str")

                design_table_data = design_table_data.round(6)

        return pd.DataFrame(design_table_data)

    def render(self) -> dash_table.DataTable:
        """Generate table."""

        data = self._get_data()

        return dash_table.DataTable(
            data=data.to_dict("records"),
            columns=[{"name": i, "id": i, "type": "text"} for i in data.columns],
            fixed_rows={"headers": True},
            style_header={
                "textAlign": "center",
                "font_family": "Roboto",
                "font_size": self.font_size,
                "fontWeight": "bold",
            },
            style_cell={
                "textAlign": "center",
                "font_family": "Roboto",
                "font_size": self.font_size,
                "whiteSpace": "normal",
            },
            style_data_conditional=[
                {
                    "if": {"column_id": "Status", "filter_query": '{Status} eq "Succeeded"'},
                    "backgroundColor": "rgb(223, 240, 208)",
                    "color": "rgb(0, 0, 0)",
                },
                {
                    "if": {"column_id": "Status", "filter_query": '{Status} eq "Not succeeded"'},
                    "backgroundColor": "rgb(254, 221, 215)",
                    "color": "rgb(0, 0, 0)",
                },
                {
                    "if": {"column_id": "Feasible", "filter_query": '{Feasible} eq "True"'},
                    "backgroundColor": "rgb(223, 240, 208)",
                    "color": "rgb(0, 0, 0)",
                },
                {
                    "if": {"column_id": "Feasible", "filter_query": '{Feasible} eq "False"'},
                    "backgroundColor": "rgb(254, 221, 215)",
                    "color": "rgb(0, 0, 0)",
                },
            ],
            style_as_list_view=True,
        )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Provide a component to handle the actor information for system and actor node types."""

from dash import dash_table
import pandas as pd


class ActorInformationTable:
    """Actor information component."""

    def __init__(self) -> None:
        """Constructor."""

        self.actor_info: dict = None
        self.actor_status_info: dict = None
        self.font_size: str = "15px"

    def _get_data(self) -> pd.DataFrame:

        actor_information_data = {
            "column_a": ["Working directory", "Processing state", "Execution duration"],
            "column_b": [None, None, None],
        }

        if self.actor_info and self.actor_status_info:
            actor_information_data["column_b"] = [
                self.actor_status_info["working dir"],
                self.actor_status_info["state"],
                "-",
            ]
            if self.actor_info["kind"] == "system":
                succeeded_designs = self.actor_status_info["succeeded_designs"]
                failed_designs = self.actor_status_info["failed_designs"]
                pending_designs = self.actor_status_info["pending_designs"]
                total_designs = self.actor_status_info["total_designs"]
                processed_designs = total_designs - pending_designs
                status = int(processed_designs / total_designs * 100)
                actor_information_data["column_a"].extend(["Processed", "Status", "Succeeded", "Not succeeded"])
                actor_information_data["column_b"].extend(
                    [f"{processed_designs} / {total_designs}", f"{status}%", succeeded_designs, failed_designs]
                )

        return pd.DataFrame(actor_information_data)

    def render(self):
        """Generate table."""

        data = self._get_data()

        return dash_table.DataTable(
            data=data.to_dict("records"),
            columns=[{"name": i, "id": i, "type": "text"} for i in data.columns],
            fixed_rows={"headers": True},
            style_header={
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": self.font_size,
                "fontWeight": "bold",
                "border": "none",
                "display": "none",
            },
            style_cell={"textAlign": "left", "font_family": "Roboto", "font_size": self.font_size, "border": "none"},
            style_as_list_view=True,
            style_table={"border": "none"},  # Hide table borders
        )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Component to display the system files view."""

import dash_bootstrap_components as dbc
from dash_extensions.enrich import dcc, html


class SystemFiles(object):
    """Design table component."""

    def __init__(self) -> None:
        """Constructor."""

        self.node_name: str = "unknown_node"
        self.omdb_file: dict = None
        self.font_size: str = "15px"

    def render(self) -> html.Div:
        """Generate view."""

        return html.Div(
            [
                dbc.Stack(
                    [
                        html.Div(
                            [
                                html.Button(
                                    "Download OMDB",
                                    disabled=True if not self.omdb_file else False,
                                    id=f"download_button_{self.node_name}_omdb",
                                ),
                                dcc.Download(id=f"download_{self.node_name}_omdb"),
                            ]
                        ),
                        html.Div(
                            [
                                html.Button(
                                    "Download CSV",
                                    disabled=True if not self.omdb_file else False,
                                    id=f"download_button_{self.node_name}_csv",
                                ),
                                dcc.Download(id=f"download_{self.node_name}_csv"),
                            ]
                        ),
                    ],
                    direction="horizontal",
                    gap=3,
                )
            ]
        )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import json
from pathlib import Path
from typing import Union


def _get_root_by_name(tcp_server_response: dict, root_name: str) -> dict:
    """Get root data by name."""

    for root in tcp_server_response["projects"]:
        if root["system"]["name"].lower() == root_name.lower():
            return root["system"]

    raise Exception(f"Unable to find root with name {root_name}.")


def _get_root_by_uid(tcp_server_response: dict, uid: str) -> dict:

    for root in tcp_server_response["projects"]:
        if root["system"]["uid"] == uid:
            return root["system"]

    raise Exception(f"Unable to find root with uid {uid}.")


def _get_node_by_uid(data: dict, uid: str) -> dict:

    pause = True

    if isinstance(data, dict):
        for key, value in data.items():
            if key == "uid" and value == uid:
                return data
            elif isinstance(value, (dict, list)):
                result = _get_node_by_uid(value, uid)
                if result is not None:
                    return result
    elif isinstance(data, list):
        for node in data:
            result = _get_node_by_uid(node, uid)
            if result is not None:
                return result

    return None


def _get_actor_hids(actor_states: dict) -> list:

    if "states" in actor_states.keys():
        return [state["hid"] for state in actor_states["states"]]
    else:
        return []


def read_optislang_logs(log_file: Union[Path, str]) -> list:

    logs = []

    log_file = Path(log_file)

    if log_file.exists():
        with open(log_file, "r") as file:
            for line in file:
                if len(line.strip().rstrip("\n").replace(" ", "")) > 0:
                    logs.append(line.strip().rstrip("\n"))
    else:
        raise FileNotFoundError(f"Unable to find log file {log_file}.")

    return logs


def test_() -> None:

    working_directory = Path(__file__).parent.absolute()
    response_file = working_directory / "get_full_project_tree_with_properties.json"

    with open(response_file) as f:
        response = json.load(f)

    test_nodes = [
        {"uid": "bcc95c97-0313-4bee-971a-a84aa2ff6dc1", "name": "{{ cookiecutter.__solution_name_slug }}"},
        {"uid": "4b3d77be-efcb-4d8f-ac49-9deeed1900e2", "name": "AMOP"},
        {"uid": "17622ac8-158f-4ae7-85f4-1156975a75e4", "name": "hook"},
        {"uid": "75746764-9e04-4bc6-a810-f8cdf0dd8bd4", "name": "ETK"},
        {"uid": "dfe29781-f507-462c-bdec-f75e726dd501", "name": "NLPQL"},
        {"uid": "d5ee9376-3885-4671-bd29-d7af3283b89c", "name": "MOP Solver"},
        {"uid": "5355d7c6-b7be-4ed0-9a35-59c4d180bbf7", "name": "Postprocessing"},
        {"uid": "70c2aadf-9a45-4098-9036-e5c27a924d16", "name": "Filter designs"},
        {"uid": "8521baa0-482c-4c93-baff-d6e8690a316d", "name": "Validator System"},
        {"uid": "0a74788b-8d54-467b-8926-5fa6f2dc1279", "name": "hook"},
        {"uid": "2892af77-59ff-4a33-b343-1f432ffac415", "name": "ETK"},
        {"uid": "c70eacf6-dc63-4844-ba6d-1d4f853e0f04", "name": "Batch Script"},
        {"uid": "c888a27c-76e3-4b15-82fa-4c95be8b5c13", "name": "Append designs"},
        {"uid": "2976c166-0ba1-4d6b-b291-3f8c1615dd3f", "name": "Validator Postprocessing"},
    ]

    # root = _get_root_by_name(response, "{{ cookiecutter.__solution_name_slug }}")
    for test_node in test_nodes:
        output = _get_node_by_uid(response, test_node["uid"])
        if output["name"] == test_node["name"]:
            print("Success")
        else:
            print("Failure")



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the summary view."""

from dash_extensions.enrich import html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.summary_view import SummaryView


def layout(
    actors_info: list, actors_status_info: dict, results_files: dict, uid: str, font_size: str = "15px"
) -> html.Div:
    """Layout of the summary view."""

    summary_view = SummaryView()

    if uid in actors_info.keys():
        summary_view.actor_info = actors_info[uid]
    if uid in actors_status_info.keys():
        summary_view.actor_status_info = actors_status_info[uid][0]
    if uid in results_files.keys():
        summary_view.result_files = results_files[uid]

    return summary_view.render()



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""General purpose functions."""

import json
from pathlib import Path
import re
from typing import Any, Iterable, Union

import dash_bootstrap_components as dbc

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.utils.constants import MONITORING_TABS


def read_system_hierarchy() -> list:

    system_hierarchy_file = (
        Path(__file__).parent.parent.parent.absolute() / "model" / "assets" / "system_hierarchy.json"
    )

    with open(system_hierarchy_file) as f:
        system_hierarchy = json.load(f)

    return system_hierarchy["system_hierarchy"]


def update_list_of_tabs(node_info: dict) -> list:

    list_of_tabs = []

    for tab_info in MONITORING_TABS:
        if (
            node_info["is_root"]
            and tab_info["is_root"]
            or not node_info["is_root"]
            and node_info["kind"] == "system"
            and tab_info["is_system"]
            or not node_info["is_root"]
            and node_info["kind"] == "actor"
            and tab_info["is_actor"]
        ):
            list_of_tabs.append(
                dbc.Tab(
                    label=tab_info["label"],
                    tab_id=tab_info["tab_id"],
                    label_style={
                        "color": "#000000",
                        "text-color": "#000000",
                    },
                    active_label_style={
                        "color": "#FFFFFF",
                        "text-color": "#000000",
                        "background-color": "#000000",
                        "border-style": "solid",
                        "border-color": "#000000",
                    },
                )
            )

    return list_of_tabs


def extract_dict_by_key(dictionaries: list, key: str, value: Any, expect_unique: bool = False) -> Union[None, dict]:
    """Given a list of dictionaries, return the ones matching specific key and value."""

    extracted_dicts = [d for d in dictionaries if d.get(key) == value]

    if expect_unique:
        if len(extracted_dicts) == 0:
            raise Exception(f"No matching dictionaries with key {key} and value {value}.")
        elif len(extracted_dicts) == 1:
            return extracted_dicts[0]
        else:
            raise Exception(f"Multiple matches found with key {key} and value {value}.")
    else:
        return extracted_dicts


def remove_key_from_dictionaries(dictionaries: list, key: str) -> list:

    for dictionary in dictionaries:
        if key in dictionary:
            del dictionary[key]

    return dictionaries


def convert_microseconds(duration: int) -> str:
    """Convert a time duration in microseconds into a string with format: WWh XXmin YYs ZZms."""

    milliseconds = duration // 1000
    microseconds = duration % 1000
    seconds = milliseconds // 1000
    milliseconds = milliseconds % 1000
    minutes = seconds // 60
    seconds = seconds % 60
    hours = minutes // 60
    minutes = minutes % 60

    return f"{int(hours):02}h {int(minutes):02}m {int(seconds):02}s {int(milliseconds):03}ms"


def sorted_nicely(l) -> Iterable:
    """Sort the given iterable in the way that humans expect.
    Reference: https://stackoverflow.com/questions/2669059/how-to-sort-alpha-numeric-set-in-python
    """

    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]
    return sorted(l, key=alphanum_key)


def sort_dict_by_ordered_keys(dictionary: dict, list_of_keys: list) -> dict:
    return {key: dictionary[key] for key in list_of_keys if key in dictionary}
